from .keys import vim
from .styles import default


class Config(object):
    """Config for GViewer

    Attributes:
        header: header content
        keys: dictionary define key mapping
        template: list of tuple for default stylesheet
    """
    def __init__(self,
                 header="General Viewer",
                 keys=vim,
                 template=default,
                 auto_scroll=False):
        self.header = header
        self.keys = keys
        self.template = template
        self.auto_scroll = auto_scroll



import unittest
import mock
import urwid

from gviewer.basic_widget import BasicWidget, FocusableText, SearchWidget, SearchableText


class TestBasicWidget(unittest.TestCase):
    def setUp(self):
        self.context = mock.Mock()
        self.controller = mock.Mock()
        self.context.config.keys = dict()

    def test_display(self):
        widget = BasicWidget()

        text = urwid.Text("This is text")
        widget.display(text)
        self.assertEqual(widget._w, text)

        text = urwid.Text("Another text")
        self.assertNotEqual(widget._w, text)
        widget.display(text)
        self.assertEqual(widget._w, text)

    def test_keypress(self):
        self.context.config.keys.update(dict(
            h="left",
            j="down",
            k="up",
            l="right"
        ))

        widget = mock.Mock()
        widget.keypress = lambda size, key: key
        widget = BasicWidget(
            controller=self.controller,
            context=self.context,
            widget=widget)

        self.assertEqual(widget.keypress((1,), "h"), "left")
        self.assertEqual(widget.keypress((1,), "left"), "left")

        self.assertEqual(widget.keypress((1,), "j"), "down")
        self.assertEqual(widget.keypress((1,), "down"), "down")

        self.assertEqual(widget.keypress((1,), "k"), "up")
        self.assertEqual(widget.keypress((1,), "up"), "up")

        self.assertEqual(widget.keypress((1,), "l"), "right")
        self.assertEqual(widget.keypress((1,), "right"), "right")

    def test_keypress_when_editing(self):
        self.context.config.keys.update(dict(
            h="left",
            j="down",
            k="up",
            l="right"
        ))

        widget = mock.Mock()
        widget.keypress = lambda size, key: key
        widget = BasicWidget(
            context=self.context,
            widget=widget)
        widget.is_editing = lambda: True

        self.assertEqual(widget.keypress((1,), "h"), "h")
        self.assertEqual(widget.keypress((1,), "j"), "j")
        self.assertEqual(widget.keypress((1,), "k"), "k")
        self.assertEqual(widget.keypress((1,), "l"), "l")


class TestFocusableText(unittest.TestCase):
    def test_contains_markup(self):
        text = FocusableText([("attr1", "text1"), ("attr2", "text2"), "text3"])

        self.assertEqual(
            [w for w in text.render((15,), False).content()],
            [[('attr1', None, 'text1'),
              ('attr2', None, 'text2'),
              (None, None, 'text3')]]
        )

        self.assertEqual(
            [w for w in text.render((15,), True).content()],
            [[(None, None, 'text1text2text3')]]
        )

    def test_plain_text(self):
        text = FocusableText("plaintext")

        self.assertEqual(
            [w for w in text.render((9,), False).content()],
            [[(None, None, 'plaintext')]]
        )

        self.assertEqual(
            [w for w in text.render((9,), True).content()],
            [[(None, None, 'plaintext')]]
        )


class TestSearchWidget(unittest.TestCase):
    def setUp(self):
        self.search_func = mock.Mock()
        self.clear_func = mock.Mock()
        self.widget = SearchWidget(
            self.search_func,
            self.clear_func
        )

    def test_invoke_search(self):
        self.widget.keypress((10,), "a")
        self.widget.keypress((10,), "b")
        self.widget.keypress((10,), "c")
        self.widget.keypress((10,), "enter")

        self.assertEqual(
            self.widget.get_keyword(),
            "abc")
        self.search_func.assert_called_with("abc")

    def test_invoke_clear(self):
        self.widget.keypress((10,), "a")
        self.widget.keypress((10,), "b")
        self.widget.keypress((10,), "c")
        self.widget.keypress((10,), "esc")

        self.assertEqual(
            self.widget.get_keyword(),
            "")
        self.clear_func.assert_called_with()


class TestSearchableText(unittest.TestCase):
    def test_found_next_one(self):
        widget = SearchableText("aaamatchbbb")
        self.assertTrue(widget.search_next("match"))
        self.assertEqual(widget.prev_index, (8, 3))
        self.assertFalse(widget.search_next("match"))

    def test_found_prev_one(self):
        widget = SearchableText("aaamatchbbb")
        self.assertTrue(widget.search_prev("match"))
        self.assertEqual(widget.prev_index, (8, 3))
        self.assertFalse(widget.search_prev("match"))

    def test_search_with_two_match(self):
        widget = SearchableText("aaamatchbbbmatchccc")
        self.assertTrue(widget.search_next("match"))
        self.assertEqual(widget.prev_index, (8, 3))
        self.assertTrue(widget.search_next("match"))
        self.assertEqual(widget.prev_index, (16, 11))
        self.assertTrue(widget.search_prev("match"))
        self.assertEqual(widget.prev_index, (8, 3))

    def test_search_in_markup(self):
        widget = SearchableText([("no-match", "aaa"), " xxx ", ("yaya", "match"), " match ", ("end", "end")])

        self.assertTrue(widget.search_next("match"))
        self.assertEqual(widget.prev_index, (3, 2))
        self.assertEqual(
            widget._w.get_text(),
            (u"aaa xxx match match end", [("no-match", 3), (None, 5), ("match", 5), (None, 7), ("end", 3)]))

        self.assertTrue(widget.search_next("match"))
        self.assertEqual(widget.prev_index, (4, 3))
        self.assertEqual(
            widget._w.get_text(),
            (u"aaa xxx match match end", [("no-match", 3), (None, 5), ("yaya", 5), (None, 1), ("match", 5), (None, 1), ("end", 3)]))

        self.assertFalse(widget.search_next("match"))
        self.assertEqual(
            widget._w.get_text(),
            urwid.Text(widget.text).get_text())

        self.assertTrue(widget.search_prev("match"))
        self.assertEqual(widget.prev_index, (4, 3))
        self.assertEqual(
            widget._w.get_text(),
            (u"aaa xxx match match end", [("no-match", 3), (None, 5), ("yaya", 5), (None, 1), ("match", 5), (None, 1), ("end", 3)]))

        self.assertTrue(widget.search_prev("match"))
        self.assertEqual(widget.prev_index, (3, 2))
        self.assertEqual(
            widget._w.get_text(),
            (u"aaa xxx match match end", [("no-match", 3), (None, 5), ("match", 5), (None, 7), ("end", 3)]))

        self.assertFalse(widget.search_prev("match"))
        self.assertEqual(
            widget._w.get_text(),
            urwid.Text(widget.text).get_text())



import traceback
import urwid
import sys

from ..basic_widget import BasicWidget


class ErrorWidget(BasicWidget):
    """Widget for displaying exception info

    Attributes:
        exc_info: sys.exc_info()
    """
    def __init__(self, **kwargs):
        exc_info = sys.exc_info()
        contents = [
            urwid.Text(s.rstrip()) for s in traceback.format_exception(
                exc_info[0], exc_info[1], exc_info[2])]
        walker = urwid.SimpleFocusListWalker(contents)
        widget = urwid.ListBox(walker)
        super(ErrorWidget, self).__init__(
            widget=widget, attr_map="error", **kwargs)

    def keypress(self, size, key):
        if key == "q":
            self.controller.back()
            return None
        return super(ErrorWidget, self).keypress(size, key)  # pragma: no cover



import urwid

from .basic_widget import BasicWidget
from .controller import Controller
from .view.summary import SummaryListWidget
from .view.error import ErrorWidget


class ParentFrame(urwid.Frame):
    """ Parent Frame to control the which widget to display

    Attributes:
        context: Context
    """
    def __init__(self, context):
        self.context = context
        self.controller = Controller(self)

        self.main = SummaryListWidget(
            context.main_context, controller=self.controller,
            context=self.context)
        self.others = {
            ctx: SummaryListWidget(ctx, controller=self.controller,
                                   context=self.context) for ctx in context.other_contexts}

        header = urwid.Text(context.config.header)
        header = urwid.AttrMap(header, "header")
        self.footer = Footer(controller=self.controller)

        self.histories = []

        super(ParentFrame, self).__init__(
            body=self.main,
            header=header,
            footer=self.footer)

    def open_view(self, widget, push_prev=True):
        """Open widget

        Args:
            widget: Widget
            push_prev: Boolean to record previous view
        """
        curr_widget = self.contents["body"][0]

        if widget is curr_widget:
            return

        if push_prev:
            self.histories.append(curr_widget)

        self.set_body(widget)

    def open_view_by_context(self, context):
        """Open view by defined context

        Args:
            context: DisplayerContext
        """
        try:
            self.open_view(self.others[context])
        except:
            self.open_error()

    def back(self):
        """Back to previous view"""
        if self.histories:
            self.set_body(self.histories.pop())
        else:
            raise urwid.ExitMainLoop()

    def open_error(self):
        """Open ErrorWidget"""
        widget = ErrorWidget(controller=self.controller, context=self.context)
        self.open_view(widget, True)

    def notify(self, message):
        """Notify message"""
        self.footer.notify(message)

    def open_edit(self, widget):
        self.footer.open_edit(widget)
        self.focus_position = "footer"

    def close_edit(self):
        self.footer.close_edit()

    def update_info(self, info):
        self.footer.update_info(info)

    def run_before_keypress(self):
        """Some UI cleanup actions to make sure GViewer in a consist state"""
        self.footer.clear_notify()


class Footer(BasicWidget):
    """Footer widget for ParentFrame"""
    def __init__(self, notification=None, helper=None, **kwargs):
        self.notification = notification or Notification(**kwargs)
        self.helper = helper or Helper()
        widgets = [
            ("pack", self.helper),
            ("pack", self.notification)
        ]
        widget = urwid.Pile(widgets)
        super(Footer, self).__init__(
            widget=widget, **kwargs)

    def notify(self, message):
        """Notify message"""
        self.notification.notify(message)

    def clear_notify(self):
        self.notification.clear()

    def open_edit(self, widget):
        self.notification.open_edit(widget)

    def close_edit(self):
        self.notification.close_edit()

    def update_info(self, info):
        self.helper.update_info(info)


class Helper(BasicWidget):
    """Helper widget contains basic help words"""
    def __init__(self, info_widget=None):
        self.info_widget = info_widget or urwid.Text("")
        widget = urwid.Text("?:help")
        widget = urwid.Padding(
            widget, "right", "pack")
        widget = urwid.Columns([self.info_widget, widget])
        super(Helper, self).__init__(
            widget=widget, attr_map="footer helper")

    def update_info(self, info):
        self.info_widget.set_text(info)


class Notification(BasicWidget):
    """Notification widget"""
    EMPTY_WIDGET = urwid.Text("")

    def __init__(self, **kwargs):
        self._edit_widget = None
        super(Notification, self).__init__(
            widget=self.EMPTY_WIDGET,
            attr_map="footer info", **kwargs)

    def is_editing(self):
        return bool(self._edit_widget)

    def selectable(self):
        return True

    def notify(self, message):
        """Notify message"""
        self.display(urwid.Text(message))

    def clear(self):
        """Clear message"""
        self.display(self._edit_widget or self.EMPTY_WIDGET)

    def open_edit(self, widget):
        self._edit_widget = widget
        self.display(widget)

    def close_edit(self):
        self._edit_widget = None
        self.display(self.EMPTY_WIDGET)



import urwid


def render_to_content(widget, size, focus=False):
    return [line for line in widget.render(size, focus).content()]


def render_widgets_to_content(widgets, size, focus=False, inline=False):
    if inline:
        widget = urwid.Columns([("pack", w) for w in widgets])
    else:
        widget = urwid.Pile([("pack", w) for w in widgets])
    return render_to_content(widget, size, focus)


def render_to_text(widget, size, focus=False):
    return widget.render(size, focus).text



import unittest

from gviewer.config import Config
from gviewer.keys import vim
from gviewer.styles import default


class TestConfig(unittest.TestCase):
    def test_default_value(self):
        config = Config()

        self.assertEqual(config.header, "General Viewer")
        self.assertEqual(config.keys, vim)
        self.assertEqual(config.template, default)



from collections import OrderedDict


keys = OrderedDict([
    ("j", "down"),
    ("k", "up"),
    ("ctrl f", "page down"),
    ("ctrl d", "page down"),
    ("ctrl b", "page up"),
    ("ctrl u", "page up")]
)



import unittest
import urwid
import mock

from .util import render_to_content, render_widgets_to_content, render_to_text
from gviewer.parent import ParentFrame, Footer, Helper, Notification
from gviewer.config import Config
from gviewer.context import Context, DisplayerContext
from gviewer.view.error import ErrorWidget
from gviewer.view.element import View, Text, Group
from gviewer.view.summary import SummaryListWidget
from gviewer.store import StaticDataStore


class TestParentFrame(unittest.TestCase):
    def setUp(self):
        self.messages = [
            ["aaa1", "aaa2", "aaa3"],
            ["bbb1", "bbb2", "bbb3"],
            ["ccc1", "ccc2", "ccc3"]
        ]
        store = StaticDataStore(self.messages)
        main_context = DisplayerContext(store, self)

        messages2 = [
            ["ddd1", "ddd2", "ddd3"],
            ["eee1", "eee2", "eee3"],
            ["fff1", "fff2", "fff3"]
        ]
        store2 = StaticDataStore(messages2)
        self.other_context = DisplayerContext(store2, self)

        context = Context(
            Config(), main_context=main_context,
            other_contexts=[self.other_context])
        self.widget = ParentFrame(context)

        store.setup()

    def summary(self, message):
        return ";".join(message)

    def get_views(self):
        return [
            ("view1", self.view),
            ("view2", self.view)
        ]

    def view(self, message):
        return View([Group("foo", [Text(m) for m in message])])

    def test_render(self):
        self.assertEqual(
            render_to_content(self.widget, (30, 10)),
            render_widgets_to_content([
                urwid.AttrMap(urwid.Text("General Viewer"), "header"),
                urwid.AttrMap(urwid.Text(";".join(self.messages[0])), "summary"),
                urwid.AttrMap(urwid.Text(";".join(self.messages[1])), "summary"),
                urwid.AttrMap(urwid.Text(";".join(self.messages[2])), "summary"),
                urwid.Text(""),
                urwid.Text(""),
                urwid.Text(""),
                urwid.Text(""),
                Footer(helper=Helper(info_widget=urwid.Text("[1/3]")))
            ], (30, 10))
        )

    def test_initial_with_main(self):
        self.assertIs(
            self.widget.contents["body"][0],
            self.widget.main
        )

    def test_back(self):
        self.widget.open_view(urwid.Text(""))
        self.widget.back()
        self.assertIs(
            self.widget.contents["body"][0],
            self.widget.main
        )

    def test_back_to_exit(self):
        with self.assertRaises(urwid.ExitMainLoop):
            self.widget.back()

    def test_open_error(self):
        try:
            raise ValueError("error")
        except:
            self.widget.open_error()

        self.assertIsInstance(
            self.widget.contents["body"][0],
            ErrorWidget
        )

    def test_notify(self):
        self.widget.notify("test")

        self.assertEqual(
            render_to_text(self.widget.footer.notification, (5,)),
            ["test "])

    def test_open_edit(self):
        self.widget.open_edit(urwid.Edit("/"))
        self.assertEqual(self.widget.focus_position, "footer")
        self.assertEqual(
            render_to_text(self.widget.footer.notification, (5,)), ["/    "])

    def test_close_edit(self):
        self.widget.open_edit(urwid.Edit("/"))
        self.widget.close_edit()
        self.assertEqual(
            render_to_text(self.widget.footer.notification, (5,)), ["     "])

    def test_run_before_keypress(self):
        self.widget.notify("test")
        self.widget.run_before_keypress()
        self.assertEqual(
            render_to_text(self.widget.footer.notification, (5,)),
            ["     "])

    def test_open_view(self):
        open_view = urwid.ListBox(urwid.SimpleFocusListWalker([]))
        self.widget.open_view(open_view)

        self.assertIs(self.widget.contents["body"][0], open_view)
        self.assertEqual(len(self.widget.histories), 1)
        self.assertIs(self.widget.histories[0], self.widget.main)

    def test_open_same_view_twice(self):
        open_view = urwid.ListBox(urwid.SimpleFocusListWalker([]))
        self.widget.open_view(open_view)

        self.assertIs(self.widget.contents["body"][0], open_view)

        self.widget.open_view(open_view)
        self.assertEqual(len(self.widget.histories), 1)

    def test_open_view_by_context(self):
        self.assertEqual(len(self.widget.others), 1)
        other_widget = self.widget.others[self.other_context]
        self.assertIsInstance(other_widget, SummaryListWidget)

        self.widget.open_view_by_context(self.other_context)
        self.assertIs(
            self.widget.contents["body"][0], other_widget)
        self.assertEqual(len(self.widget.histories), 1)
        self.assertIs(self.widget.histories[0], self.widget.main)

    def test_open_view_by_context_failed(self):
        self.widget.open_view_by_context(DisplayerContext(None, None))
        self.assertIsInstance(
            self.widget.contents["body"][0], ErrorWidget)
        self.assertEqual(len(self.widget.histories), 1)
        self.assertIs(self.widget.histories[0], self.widget.main)


class TestFooter(unittest.TestCase):
    def setUp(self):
        self.helper = mock.Mock()
        self.notification = mock.Mock()
        self.widget = Footer(helper=self.helper, notification=self.notification)

    def test_render(self):
        widget = Footer()
        self.assertEqual(
            render_to_content(widget, (20, 2)),
            render_widgets_to_content(
                [Helper(), Notification()],
                (20, 2)))

    def test_notify(self):
        self.widget.notify("test")
        self.notification.notify.assert_called_with("test")

    def test_clear_notify(self):
        self.widget.clear_notify()
        self.notification.clear.assert_called_with()

    def test_open_edit(self):
        w = urwid.Edit()
        self.widget.open_edit(w)
        self.notification.open_edit.assert_called_with(w)

    def test_close_edit(self):
        self.widget.close_edit()
        self.notification.close_edit.assert_called_with()


class TestNotification(unittest.TestCase):
    def setUp(self):
        self.widget = Notification()

    def test_render_default(self):
        self.assertEqual(
            render_to_text(self.widget, (5,)),
            ["     "])

    def test_notify(self):
        self.widget.notify("test")
        self.assertEqual(
            render_to_text(self.widget, (5,)),
            ["test "])

    def test_clear_notify(self):
        self.widget.notify("test")
        self.widget.clear()
        self.assertEqual(
            render_to_text(self.widget, (5,)),
            ["     "])

    def test_open_edit(self):
        w = urwid.Edit("/")
        self.widget.open_edit(w)

        self.assertIs(self.widget._edit_widget, w)
        self.assertTrue(self.widget.is_editing())
        self.assertEqual(
            render_to_text(self.widget, (5,)),
            ["/    "])

    def test_notify_on_edit(self):
        self.widget.open_edit(urwid.Edit("/"))
        self.widget.notify("hahaha")

        self.assertEqual(
            render_to_text(self.widget, (6,)),
            ["hahaha"])

        self.widget.clear()
        self.assertEqual(
            render_to_text(self.widget, (6,)),
            ["/     "])

    def test_close_edit(self):
        self.widget.open_edit(urwid.Edit("/"))
        self.widget.close_edit()
        self.assertFalse(self.widget.is_editing())
        self.assertIsNone(self.widget._edit_widget)



from gviewer import GViewer
from store import BaseDataStore, StaticDataStore, AsyncDataStore
from displayer import BaseDisplayer
from config import Config
from view.element import Text, Prop, Group, PropsGroup, View
from action import Actions
from context import DisplayerContext



import unittest
import mock

from gviewer.store import StaticDataStore, AsyncDataStore


class TestStaticDataStore(unittest.TestCase):
    def setUp(self):
        self.walker = mock.Mock()

        self.data_store = StaticDataStore([
            "message 1", "message 2"])
        self.data_store.register(self.walker)

    def test_setup(self):
        self.assertEqual(len(self.data_store.walkers), 1)

        self.data_store.setup()
        self.walker.recv.assert_has_calls(
            [mock.call("message 1"), mock.call("message 2")], any_order=True)

    def test_register_new_walker(self):
        walker2 = mock.Mock()
        self.data_store.register(walker2)

        self.assertEqual(len(self.data_store.walkers), 2)

        self.data_store.setup()
        self.walker.recv.assert_has_calls(
            [mock.call("message 1"), mock.call("message 2")], any_order=True)
        walker2.recv.assert_has_calls(
            [mock.call("message 1"), mock.call("message 2")], any_order=True)

    def test_unregister(self):
        self.data_store.unregister(self.walker)
        self.assertEqual(len(self.data_store.walkers), 0)


class TestAsyncDataStore(unittest.TestCase):
    def setUp(self):
        self.walker = mock.Mock()

        self.register_func = mock.Mock()

        self.data_store = AsyncDataStore(self.register_func)
        self.data_store.register(self.walker)

    def test_setup(self):
        self.data_store.setup()
        self.register_func.assert_called_with(
            self.data_store.on_message)



import unittest
import mock

from ..util import render_to_content
from gviewer.view.error import ErrorWidget


class TestErrorWidget(unittest.TestCase):
    def setUp(self):
        self.controller = mock.Mock()
        self.context = mock.Mock()

    def test_render(self):
        try:
            raise ValueError("wrong value")
        except:
            widget = ErrorWidget(
                controller=self.controller,
                context=self.context
            )

        contents = render_to_content(widget, (200, 4))
        self.assertEqual(len(contents), 4)

        self.assertEqual(contents[0][0][0], "error")
        self.assertEqual(contents[0][0][2].rstrip(), "Traceback (most recent call last):")

        self.assertEqual(contents[1][0][0], "error")
        self.assertTrue("test_error.py" in contents[1][0][2])
        self.assertTrue("test_render" in contents[1][0][2])

        self.assertEqual(contents[2][0][0], "error")
        self.assertEqual(contents[2][0][2].strip(), "raise ValueError(\"wrong value\")")

        self.assertEqual(contents[3][0][0], "error")
        self.assertEqual(contents[3][0][2].strip(), "ValueError: wrong value")

    def test_keypress_q(self):
        try:
            raise ValueError("wrong value")
        except:
            widget = ErrorWidget(
                controller=self.controller,
                context=self.context
            )
        self.assertEqual(widget.keypress(None, "q"), None)
        self.controller.back.assert_called_with()



class Controller(object):
    """Controller provide UI interaction API

    Attributes:
        parent: ParentFrame instance
    """
    def __init__(self, parent):
        self.parent = parent

    def open_view(self, widget, push_prev=True):
        """Open view

        Args:
            widget: a urwid Widget implementation
            push_prev: bool defined that should push previous widget
                       into history
        """
        self.parent.open_view(widget, push_prev=push_prev)

    def open_view_by_context(self, context):
        """Open view by defined context

        Args:
            context: DisplayerContext
        """
        self.parent.open_view_by_context(context)

    def notify(self, message):
        """Notify a message"""
        self.parent.notify(message)

    def open_edit(self, widget):
        """Open edit box in footer"""
        self.parent.open_edit(widget)

    def close_edit(self):
        """Close edit box"""
        self.parent.close_edit()

    def open_error(self):
        """Open error stacktrace view"""
        self.parent.open_error()

    def back(self):
        """Back to previous view"""
        self.parent.back()

    def _update_info(self, info):
        self.parent.update_info(info)

    def _focus_body(self):
        self.parent.focus_position = "body"

    def _run_before_keypress(self):
        self.parent.run_before_keypress()



import unittest
import urwid
import mock

from ..util import render_to_content, render_widgets_to_content
from gviewer.view.summary import (
    SummaryItemWidget, SummaryListWalker,
    FilterSummaryListWalker, SummaryListWidget)
from gviewer.view.summary import _verify_keys
from gviewer.view.detail import DetailWidget
from gviewer.view.element import View
from gviewer.action import Actions
from gviewer.store import StaticDataStore
from gviewer.context import DisplayerContext
from gviewer.config import Config
from gviewer.displayer import BaseDisplayer


class TestSummaryItemWidget(unittest.TestCase):
    def setUp(self):
        self.controller = mock.Mock()
        self.controller.open_view = mock.Mock(side_effect=self._open_view)

        self.context = mock.Mock()
        self.context.config.keys = dict()

        self.action_a = mock.Mock()
        self.displayer_context = DisplayerContext(
            None, self, Actions([("a", "aaaa", self.action_a)]))

        self.widget = SummaryItemWidget(
            "message", "summary", self.displayer_context, controller=self.controller,
            context=self.context)

        self.new_widget = None

    def get_views(self):
        return [("view", self.view1)]

    def view1(self, message):
        return View([])

    def _open_view(self, widget, **kwargs):
        self.new_widget = widget

    def test_render(self):
        self.assertEqual(
            render_to_content(self.widget, (7,)),
            render_widgets_to_content(
                [urwid.AttrMap(urwid.Text("summary"), "summary")],
                (7, 1))
        )

    def test_keypress_enter(self):
        self.assertIsNone(self.widget.keypress(None, "enter"))
        self.assertIsNotNone(self.new_widget)
        self.assertIsInstance(self.new_widget, DetailWidget)

    def test_keypress_custom_action(self):
        self.assertIsNone(self.widget.keypress(None, "a"))
        self.action_a.assert_called_with(self.controller, "message", self.widget)

    def test_set_title(self):
        self.widget.set_title("hahaha")
        self.assertEqual(self.widget.get_title_as_plain_text(), "hahaha")


class TestSummaryListWalker(unittest.TestCase):
    def setUp(self):
        self.controller = mock.Mock()
        self.controller.open_error = mock.Mock(
            side_effect=self._open_error)

        self.context = mock.Mock()
        self.context.config.keys = dict()

        self.displayer_context = mock.Mock()
        self.displayer_context.displayer.summary = mock.Mock(
            side_effect=lambda m: m)
        self.displayer_context.displayer.match = mock.Mock(
            side_effect=lambda k, m, s: k in s)
        self.on_receive = mock.Mock()

        self.error = False

        self.walker = SummaryListWalker(
            controller=self.controller, context=self.context,
            displayer_context=self.displayer_context,
            on_receive=self.on_receive)

    def _open_error(self, *args, **kwargs):
        self.error = True

    def test_msg_listener_register(self):
        self.displayer_context.store.register.assert_called_once_with(
            self.walker)

    def test_recv(self):
        self.displayer_context.displayer.summary = mock.Mock(return_value="message")

        self.walker.recv("new message")
        self.assertEqual(len(self.walker), 1)
        self.assertFalse(self.error)

        self.displayer_context.displayer.summary.assert_called_once_with("new message")

    def test_recv_failed(self):
        self.displayer_context.displayer.summary = mock.Mock(
            side_effect=ValueError("failed"))

        self.walker.recv("new message")
        self.assertEqual(len(self.walker), 0)
        self.assertTrue(self.error)


class TestFilterSummaryListWalker(unittest.TestCase):
    def setUp(self):
        self.controller = mock.Mock()
        self.controller.open_error = mock.Mock(
            side_effect=self._open_error)

        self.context = mock.Mock()
        self.context.config.keys = dict()

        self.displayer_context = mock.Mock()
        self.displayer_context.displayer = BaseDisplayer()
        self.on_receive = mock.Mock()
        self.error = False

        self.original_walker = SummaryListWalker(
            content=[
                SummaryItemWidget(
                    "message 1", "summary 1",
                    self.displayer_context,
                    controller=self.controller,
                    context=self.context),
                SummaryItemWidget(
                    "message 2", "summary 2",
                    self.displayer_context,
                    controller=self.controller,
                    context=self.context)
            ],
            controller=self.controller,
            context=self.context,
            displayer_context=self.displayer_context,
            on_receive=self.on_receive
        )

    def _open_error(self, *args, **kwargs):
        self.error = True

    def test_construct(self):
        walker = FilterSummaryListWalker(
            self.original_walker, "summary 1")
        self.assertEqual(len(walker), 1)

    def test_recv_with_match(self):
        walker = FilterSummaryListWalker(
            self.original_walker, "summary 1")
        walker.recv("summary 1111")
        self.assertEqual(len(walker), 2)

    def test_recv_with_not_match(self):
        walker = FilterSummaryListWalker(
            self.original_walker, "summary 1")
        walker.recv("summary")
        self.assertEqual(len(walker), 1)

    def test_recv_failed(self):
        self.displayer_context.displayer.summary = mock.Mock(side_effect=ValueError("failed"))

        walker = FilterSummaryListWalker(
            self.original_walker, "summary 1")

        walker.recv("summary 1")
        self.assertEqual(len(walker), 1)
        self.assertTrue(self.error)

    def test_close(self):
        walker = FilterSummaryListWalker(
            self.original_walker, "summary 1")
        walker.close()
        self.displayer_context.store.unregister.assert_called_once_with(walker)

    def test_del_one_item(self):
        walker = FilterSummaryListWalker(
            self.original_walker, "")
        del walker[0]
        self.assertEqual(len(walker), 1)
        self.assertEqual(len(self.original_walker), 1)
        self.assertIs(walker[0], self.original_walker[0])

    def test_del_all_items(self):
        walker = FilterSummaryListWalker(
            self.original_walker, "summary 1")
        del walker[:]
        self.assertEqual(len(walker), 0)
        self.assertEqual(len(self.original_walker), 1)


class TestSummaryListWidget(unittest.TestCase):
    def setUp(self):
        self.controller = mock.Mock()

        self.displayer_context = DisplayerContext(
            StaticDataStore(["summary 1", "summary 2"]),
            BaseDisplayer())

        self.context = mock.Mock()
        self.context.config = Config()

        self.widget = SummaryListWidget(
            self.displayer_context,
            controller=self.controller,
            context=self.context)

        self.displayer_context.store.setup()

    def test_render(self):
        self.assertEqual(
            render_to_content(self.widget, (9, 2)),
            render_widgets_to_content([
                urwid.AttrMap(urwid.Text("summary 1"), "summary"),
                urwid.AttrMap(urwid.Text("summary 2"), "summary")
            ], (9, 2))
        )

    def test_open_search(self):
        self.widget._open_search()
        self.controller.open_edit.assert_called_with(self.widget.search_widget)

    def test_filter(self):
        self.widget._filter("summary 1")
        self.assertIsInstance(self.widget.current_walker, FilterSummaryListWalker)
        self.assertEqual(len(self.widget.current_walker), 1)

    def test_filter_with_empty(self):
        self.widget._filter("")
        self.assertIs(
            self.widget.current_walker,
            self.widget.base_walker)

    def test_keypress_open_search(self):
        self.widget.keypress(None, "/")
        self.controller.open_edit.assert_called_with(self.widget.search_widget)

    def test_keypress_clear_search(self):
        self.widget._filter("test")
        self.assertEqual(len(self.widget.current_walker), 0)

        self.widget.keypress((0,), "q")
        self.assertIs(
            self.widget.current_walker,
            self.widget.base_walker)
        self.assertEqual(len(self.widget.current_walker), 2)

    def test_keypress_back(self):
        self.widget.keypress((0,), "q")
        self.controller.back.assert_called_with()

    def test_keypress_open_help(self):
        self.assertIsNone(self.widget.keypress((0, 0), "?"))
        self.controller.open_view.assert_called_with(self.widget.help_widget)

    def test_keypress_bottom_and_top(self):
        self.widget.keypress((10, 10), "G")
        self.assertEqual(self.widget._w.focus_position, 1)
        self.controller._update_info.assert_called_with("[2/2]")

        self.widget.keypress((10, 10), "g")
        self.assertEqual(self.widget._w.focus_position, 0)
        self.controller._update_info.assert_called_with("[1/2]")

    def test_keypress_bottom_and_top_when_search(self):
        self.widget._filter("summary")
        self.widget.keypress((10, 10), "G")
        self.assertEqual(self.widget._w.focus_position, 1)
        self.widget.keypress((10, 10), "g")
        self.assertEqual(self.widget._w.focus_position, 0)

    def test_keypress_clear_item(self):
        self.widget.keypress((10, 10), "x")
        self.assertEqual(len(self.widget.current_walker), 1)
        self.controller._update_info.assert_called_with("[1/1]")

    def test_keypress_clear_items(self):
        self.widget.keypress((10, 10), "X")
        self.assertEqual(len(self.widget.current_walker), 0)
        self.controller._update_info.assert_called_with("[0/0]")

    def test_keypress_up_and_down(self):
        self.widget.keypress((10, 10), "down")
        self.assertEqual(self.widget._w.focus_position, 1)
        self.controller._update_info.assert_called_with("[2/2]")

        self.widget.keypress((10, 10), "up")
        self.assertEqual(self.widget._w.focus_position, 0)
        self.controller._update_info.assert_called_with("[1/2]")

    def test_auto_scroll(self):
        displayer_context = DisplayerContext(
            StaticDataStore(["summary 1", "summary 2"]),
            BaseDisplayer())
        self.context.config = Config(auto_scroll=True)
        widget = SummaryListWidget(
            displayer_context,
            controller=self.controller,
            context=self.context)
        displayer_context.store.setup()

        self.assertEqual(widget._w.focus_position, 1)
        widget.current_walker.recv("summary 3")
        self.assertEqual(widget._w.focus_position, 2)

        widget._w.set_focus(0)
        widget.current_walker.recv("summary 4")
        self.assertEqual(widget._w.focus_position, 0)


class TestSummary(unittest.TestCase):
    def test_verify_keys(self):
        _verify_keys(Actions([("p", "pppp", None)]))

    def test_verify_keys_failed(self):
        with self.assertRaises(ValueError):
            _verify_keys(Actions([("/", "search", None)]))



from .action import Actions


class Context(object):
    """GViewer context

    Attributes:
        config: Config instance
        main_displayer_context: DisplayerContext instance for main view
    """
    def __init__(self, config, main_context, other_contexts=None):
        self.config = config
        self.main_context = main_context
        self.other_contexts = other_contexts or []


class DisplayerContext(object):
    """Context defined for how ui display

    Attributes:
        store: DataStore instance
        displayer: BaseDisplayer implementation instance
        actions: Actions instance
    """
    def __init__(self, store, displayer, actions=None):
        self.store = store
        self.displayer = displayer
        self.actions = actions or Actions()



import urwid
from urwid.util import decompose_tagmarkup


class BasicWidget(urwid.WidgetWrap):
    """ Basic widget for GViewer

    Define the basic attr and basic behavior how GViewer widget to display or act

    Attributes:
        controller: a Controller instance
        context: a Context instance
        widget: a urwid Widget
        attr_map: non-focus attribute
        focus_map: focus attribute
    """
    def __init__(self, controller=None, context=None, widget=None,
                 attr_map=None, focus_map=None):
        widget = widget or urwid.Text("")
        if attr_map:
            widget = urwid.AttrMap(widget, attr_map, focus_map=focus_map)
        super(BasicWidget, self).__init__(widget)
        self.controller = controller
        self.context = context

    def display(self, widget):
        """ Display the widget

        Args:
            widget: urwid Widget
        """
        if isinstance(self._w, urwid.AttrMap):
            self._w.original_widget = widget
        else:
            self._w = widget

    def keypress(self, size, key):
        """ Define default keypress action for GViewer """
        if self.controller:
            self.controller._run_before_keypress()

        if (not self.is_editing() and self.context and
                key in self.context.config.keys):
            return super(BasicWidget, self).keypress(size, self.context.config.keys[key])
        try:
            return super(BasicWidget, self).keypress(size, key)
        except AttributeError:  # pragma: no cover
            return key

    def is_editing(self):
        """ Check that current state is editing or not

        Override this method if that widget had Edit widget in its content
        and return True if the focus widget is the Edit widget

        Returns:
            True if in editing, False if not in editing
        """
        return False


class FocusableText(BasicWidget):
    """ Text widget that will highlight correctly

    Attributes:
        text_markup: urwid Text Markup instance
    """
    def __init__(self, text_markup, **kwargs):
        widget = urwid.Text(text_markup)
        super(FocusableText, self).__init__(
            widget=widget, **kwargs)
        self.text_markup = text_markup

    def render(self, size, focus=False):
        """ Override original render function

        Display the widget in different way depend on that the widget is focus or not
        """
        if focus:
            plain_text = self.get_plain_text()
            self.display(urwid.Text(plain_text))
        else:
            self.display(urwid.Text(self.text_markup))
        return super(FocusableText, self).render(size, focus)

    def get_plain_text(self):
        """ Retrieve the plain text from text_markup """
        if isinstance(self.text_markup, str) or \
           isinstance(self.text_markup, unicode):
            return self.text_markup
        text, _ = decompose_tagmarkup(self.text_markup)
        return text

    def selectable(self):
        return True


class SearchWidget(BasicWidget):
    """ Edit widget to handle searching

    Attributes:
        search_func: a callback function that will be called when search is invoke
        clear_func: a callback function that will be called when search is cancelled
    """
    def __init__(self, search_func, clear_func, **kwargs):
        super(SearchWidget, self).__init__(
            widget=urwid.Edit("/"), **kwargs)
        self.search_func = search_func
        self.clear_func = clear_func

    def clear(self):
        self.display(urwid.Edit("/"))

    def get_keyword(self):
        return self._w.edit_text

    def keypress(self, size, key):
        if key == "enter" and self._w.edit_text:
            self.search_func(self._w.edit_text)
            return None
        if key == "esc" or (key == "enter" and not self._w.edit_text):
            self.clear()
            self.clear_func()
            return None
        return super(SearchWidget, self).keypress(size, key)


class SearchableText(BasicWidget):
    """ Searchable text let its content could be search and hightlight

    Attributes:
        text: str or unicode or text_markup
    """
    def __init__(self, text, **kwargs):
        if isinstance(text, tuple):  # NOTE: not accept tuple type markup
            raise NotImplementedError("not accept tuple type markup")

        super(SearchableText, self).__init__(
            widget=urwid.Text(text), **kwargs)

        self.text = text
        self.prev_index = (0, len(text))

    def search_next(self, keyword):
        prev_index = self.prev_index[0]
        if isinstance(keyword, str):
            keyword = keyword.decode("utf8")

        if isinstance(self.text, (str, unicode)):
            if keyword in self.text[prev_index:]:
                start_index = self.text[prev_index:].index(keyword) + prev_index
                self._handle_match_plain_text(keyword, start_index)
                return True
        else:
            for index in range(prev_index, len(self.text)):
                plain_text, _ = decompose_tagmarkup(self.text[index])
                plain_text = plain_text.decode("utf8")
                if keyword in plain_text:
                    self._handle_match_markup(keyword, plain_text, index)
                    return True

        self.clear()
        return False

    def search_prev(self, keyword):
        prev_index = self.prev_index[1]
        if isinstance(keyword, str):
            keyword = keyword.decode("utf8")

        if isinstance(self.text, (str, unicode)):
            if keyword in self.text[:prev_index]:
                start_index = self.text[:prev_index].rindex(keyword)
                self._handle_match_plain_text(keyword, start_index)
                return True
        else:
            for index in reversed(range(0, prev_index)):
                plain_text, _ = decompose_tagmarkup(self.text[index])
                plain_text = plain_text.decode("utf8")
                if keyword in plain_text:
                    self._handle_match_markup(keyword, plain_text, index)
                    return True

        self.clear()
        return False

    def _handle_match_plain_text(self, keyword, start_index):
        end_index = start_index + len(keyword)
        self.display(urwid.Text([
            self.text[:start_index], ("match", keyword), self.text[end_index:]]))
        self.prev_index = (end_index, start_index)

    def _handle_match_markup(self, keyword, plain_text, index):
            match_index = plain_text.index(keyword)
            match_end_index = match_index + len(keyword)
            match_markup = []
            if plain_text[:match_index]:
                match_markup.append(plain_text[:match_index])
            match_markup.append(("match", keyword))
            if plain_text[match_end_index:]:
                match_markup.append(plain_text[match_end_index:])
            new_markup = self.text[:index] + match_markup + self.text[index + 1:]
            self.display(urwid.Text(new_markup))
            self.prev_index = (index + 1, index)

    def get_plain_text(self):
        return decompose_tagmarkup(self.text)[0].decode("utf8")

    def clear(self):
        self.prev_index = (0, len(self.text))
        self.display(urwid.Text(self.text))



import unittest
import mock

from gviewer.action import Action, Actions


class TestAction(unittest.TestCase):
    def test_call(self):
        function = mock.Mock()
        action = Action("desc", function)

        action.__call__()
        function.assert_called_with()

        action.__call__("aaa")
        function.assert_called_with("aaa")

        action.__call__(aaa="bbb")
        function.assert_called_with(aaa="bbb")


class TestActions(unittest.TestCase):
    def setUp(self):
        self.func_a = mock.Mock()
        self.func_b = mock.Mock()
        self.actions = Actions([
            ("a", "haha", self.func_a),
            ("b", "yaya", self.func_b)])

    def test_iter(self):
        action_list = [a for a in self.actions]
        self.assertEqual(len(action_list), 2)
        self.assertEqual(
            action_list[0],
            ("a", "haha", self.func_a))
        self.assertEqual(
            action_list[1],
            ("b", "yaya", self.func_b))

    def test_contains(self):
        self.assertTrue("a" in self.actions)
        self.assertTrue("b" in self.actions)
        self.assertFalse("c" in self.actions)

    def test_getitem(self):
        action = self.actions["a"]
        self.assertIsInstance(action, Action)
        self.assertEqual(action.desc, "haha")



import unittest
import urwid
import mock

from ..util import render_to_content, render_widgets_to_content
from gviewer.basic_widget import SearchableText
from gviewer.view.element import (
    Text, Prop, Group, PropsGroup, View,
    TitleWidget, ContentWidget, EmptyLine)


class TestText(unittest.TestCase):
    def test_widget(self):
        widget = Text("content").widget(None)
        self.assertTrue(isinstance(widget, SearchableText))
        self.assertEqual(
            widget.get_plain_text(),
            u"content")

    def test_to_text(self):
        self.assertEqual(
            str(Text(b"content")), "content")
        self.assertEqual(
            str(Text(u"content")), "content")

    def test_markup_to_text(self):
        self.assertEqual(
            str(Text([("aaa", "bbb"), "ccc"])), "bbbccc")


class TestProp(unittest.TestCase):
    def test_widget(self):
        widget = Prop("key", "value").widget(None)
        self.assertIsInstance(widget, SearchableText)
        self.assertEqual(
            widget.text,
            [('view-item key', u'key: '),
             ('view-item value', u'value')])

    def test_widget_with_padding(self):
        prop = Prop("key", "value")
        prop.max_key_length = 5
        self.assertEqual(
            prop.widget(None).text,
            [('view-item key', u'key   : '),
             ('view-item value', u'value')])

    def test_to_text(self):
        self.assertEqual(
            str(Prop("key", "value")),
            "key: value")

    def test_to_text_with_padding(self):
        prop = Prop("key", "value")
        prop.max_key_length = 5
        self.assertEqual(
            str(prop),
            "key   : value")


class TestGroup(unittest.TestCase):
    def test_widget_with_title(self):
        widgets = Group(
            "title",
            items=[
                Text("first line"),
                Text("second line")]
        ).widgets(None)

        self.assertEqual(len(widgets), 3)
        self.assertTrue(isinstance(widgets[0], TitleWidget))
        self.assertTrue(isinstance(widgets[1], SearchableText))
        self.assertTrue(isinstance(widgets[2], SearchableText))

    def test_widget_without_title(self):
        widgets = Group(
            "title",
            items=[
                Text("first line"),
                Text("second line")],
            show_title=False).widgets(None)

        self.assertEqual(len(widgets), 2)
        self.assertTrue(isinstance(widgets[0], SearchableText))
        self.assertTrue(isinstance(widgets[1], SearchableText))

    def test_to_text(self):
        self.assertEqual(
            str(Group("title", items=[Text("first line"), Text("second line")],
                show_title=True)),
            u"title\nfirst line\nsecond line")

        self.assertEqual(
            str(Group("title", items=[Text("first line"), Text("second line")],
                show_title=False)),
            u"first line\nsecond line")


class TestPropsGroup(unittest.TestCase):
    def test_calc_key_length(self):
        props_group = PropsGroup(
            "title",
            items=[
                Prop("key", "value"),
                Prop("longkey", "value"),
                Prop("k", "value")]
        )
        self.assertEqual(props_group.items[0].max_key_length, 7)
        self.assertEqual(props_group.items[1].max_key_length, 7)
        self.assertEqual(props_group.items[2].max_key_length, 7)


class TestView(unittest.TestCase):
    def test_widget(self):
        view = View([
            Group("group1", [Text("content1")]),
            Group("group2", [Text("content2")])]
        )
        widget = view.widget(None)
        self.assertTrue(isinstance(widget, ContentWidget))

        contents = widget._w.body
        self.assertEqual(len(contents), 6)
        self.assertIsInstance(contents[0], TitleWidget)
        self.assertIsInstance(contents[1], SearchableText)
        self.assertIsInstance(contents[2], EmptyLine)
        self.assertIsInstance(contents[3], TitleWidget)
        self.assertIsInstance(contents[4], SearchableText)
        self.assertIsInstance(contents[5], EmptyLine)

    def test_empty_widget(self):
        view = View([])
        contents = view.widget(None)._w.body
        self.assertEqual(len(contents), 1)
        self.assertIsInstance(contents[0], EmptyLine)

    def test_actions(self):
        action = mock.Mock()
        controller = mock.Mock()
        context = mock.Mock()
        context.config.keys = dict()

        view = View([
            Group("group1", [Text("content1")]),
            Group("group2", [Text("content2")])],
            actions=dict(a=action)
        )
        widget = view.widget("message", controller=controller, context=context)
        widget.keypress((0, 0), "a")
        action.assert_called_with(controller, "message")

        self.assertEqual(
            widget.keypress((0, 0), "b"),
            "b")

    def test_to_text(self):
        view = View([
            Group("group1", [Text("content1")]),
            Group("group2", [Text("content2")])]
        )
        self.assertEqual(
            str(view),
            u"group1\ncontent1\n\ngroup2\ncontent2\n")


class TestTitleWidget(unittest.TestCase):
    def test_render(self):
        widget = TitleWidget("content")
        contents = [c for c in widget.render((7,)).content()]
        self.assertEqual(len(contents), 1)
        self.assertEqual(len(contents[0]), 1)
        self.assertEqual(contents[0][0], ("view-title", None, "content"))


class TestContentWidget(unittest.TestCase):
    def test_search(self):
        widget = ContentWidget([
            SearchableText("this is aaa bbb"),
            SearchableText("this is ccc ddd"),
            SearchableText("this is eee aaa")
        ], None)

        no_match = render_widgets_to_content([
            urwid.Text("this is aaa bbb"),
            urwid.Text("this is ccc ddd"),
            urwid.Text("this is eee aaa")],
            (15, 3)
        )
        first_match = render_widgets_to_content([
            urwid.Text(["this is ", ("match", "aaa"), " bbb"]),
            urwid.Text("this is ccc ddd"),
            urwid.Text("this is eee aaa")],
            (15, 3)
        )
        second_match = render_widgets_to_content([
            urwid.Text("this is aaa bbb"),
            urwid.Text("this is ccc ddd"),
            urwid.Text(["this is eee ", ("match", "aaa")])],
            (15, 3)
        )

        self.assertEqual(
            render_to_content(widget, (15, 3)),
            no_match
        )

        widget.search_next("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            first_match
        )

        widget.search_next("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            second_match
        )

        widget.search_next("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            no_match
        )

        widget.search_prev("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            second_match
        )

        widget.search_prev("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            first_match
        )

        widget.search_prev("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            no_match
        )

    def test_search_next_before_move(self):
        widget = ContentWidget([
            SearchableText("this is aaa bbb"),
            SearchableText("this is ccc ddd"),
            SearchableText("this is eee aaa")
        ], None)

        no_match = render_widgets_to_content([
            urwid.Text("this is aaa bbb"),
            urwid.Text("this is ccc ddd"),
            urwid.Text("this is eee aaa")],
            (15, 3)
        )
        first_match = render_widgets_to_content([
            urwid.Text(["this is ", ("match", "aaa"), " bbb"]),
            urwid.Text("this is ccc ddd"),
            urwid.Text("this is eee aaa")],
            (15, 3)
        )
        second_match = render_widgets_to_content([
            urwid.Text("this is aaa bbb"),
            urwid.Text("this is ccc ddd"),
            urwid.Text(["this is eee ", ("match", "aaa")])],
            (15, 3)
        )

        self.assertEqual(
            render_to_content(widget, (15, 3)),
            no_match
        )

        widget.search_next("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            first_match
        )

        widget._w.focus_position = 2

        widget.search_next("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            second_match
        )

    def test_search_prev_before_move(self):
        widget = ContentWidget([
            SearchableText("this is aaa bbb"),
            SearchableText("this is ccc ddd"),
            SearchableText("this is eee aaa")
        ], None)

        no_match = render_widgets_to_content([
            urwid.Text("this is aaa bbb"),
            urwid.Text("this is ccc ddd"),
            urwid.Text("this is eee aaa")],
            (15, 3)
        )
        first_match = render_widgets_to_content([
            urwid.Text(["this is ", ("match", "aaa"), " bbb"]),
            urwid.Text("this is ccc ddd"),
            urwid.Text("this is eee aaa")],
            (15, 3)
        )
        second_match = render_widgets_to_content([
            urwid.Text("this is aaa bbb"),
            urwid.Text("this is ccc ddd"),
            urwid.Text(["this is eee ", ("match", "aaa")])],
            (15, 3)
        )

        self.assertEqual(
            render_to_content(widget, (15, 3)),
            no_match
        )

        widget.search_next("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            first_match
        )

        widget._w.focus_position = 2

        widget.search_prev("aaa")
        self.assertEqual(
            render_to_content(widget, (15, 3)),
            second_match
        )



try:
    from pygments import token
except:  # pragma: no cover
    pygments_style = []
else:
    pygments_style = [
        (token.Token.Text, "white", "black"),
        (token.Token.Name, "white", "black"),
        (token.Token.Name.Builtin, "light magenta", "black"),
        (token.Token.Name.Function, "light magenta", "black"),
        (token.Token.Name.Class, "light magenta", "black"),
        (token.Token.Comment, "dark gray", "black"),
        (token.Token.Literal, "brown", "black"),
        (token.Token.Literal.String, "brown", "black"),
        (token.Token.Literal.String.Number, "light cyan", "black"),
        (token.Token.Literal.String.Double, "light cyan", "black"),
        (token.Token.Literal.String.Escape, "light cyan", "black"),
        (token.Token.Literal.Number.Integer, "brown", "black"),
        (token.Token.Operator, "white", "black"),
        (token.Token.Error, "white", "black"),
        (token.Token.Other, "white", "black"),
        (token.Token.Keyword, "dark cyan", "black"),
        (token.Token.Keyword.Namespace, "light green", "black"),
        (token.Token.Generic, "white", "black")
    ]

palette = [
    ("header", "white", "dark green", "bold"),
    ("footer helper", "white", "dark blue", "bold"),
    ("footer info", "white", "black"),
    ("summary", "white", "black"),
    ("summary focus", "black", "light gray"),
    ("view-item", "white", "black"),
    ("view-item key", "light cyan", "black"),
    ("view-item value", "white", "black"),
    ("view-title", "yellow", "black", "bold"),
    ("tabs", "white", "black", "bold"),
    ("tabs focus", "white", "dark blue", "bold"),
    ("match", "black", "light gray"),
    ("help title", "white", "black", "bold"),
    ("help key", "light cyan", "black"),
    ("help value", "white", "black")
] + pygments_style



import unittest
from pygments import lexers, token

from gviewer.util import pygmentize, _join


class TestUtil(unittest.TestCase):
    def test_pygmentize(self):
        python_content = """
        import unittest

        class Pygmentize(object):
            pass"""
        result = pygmentize(python_content, lexers.PythonLexer())

        self.assertEqual(len(result), 4)

        self.assertIn(
            (token.Token.Keyword.Namespace, u'import'),
            result[0])
        self.assertIn(
            (token.Token.Name.Namespace, u'unittest'),
            result[0])

        self.assertEqual(result[1], u"")

        self.assertIn(
            (token.Token.Keyword, u'class'),
            result[2])
        self.assertIn(
            (token.Token.Name.Class, u'Pygmentize'),
            result[2])

        self.assertIn(
            (token.Token.Keyword, u'pass'),
            result[3])

    def test_join(self):
        result = _join([("aaa", "bbb"), ("ccc", "ddd")], "\n")
        self.assertEqual(len(result), 1)
        self.assertEqual(
            result[0], [("aaa", "bbb"), ("ccc", "ddd")])



import urwid
from urwid.util import decompose_tagmarkup

from ..basic_widget import BasicWidget, SearchableText
from ..action import Actions


class Base(object):  # pragma: no cover
    """Abstract class for view displayer eleemnt"""
    def widget(self, message, controller=None, context=None):
        raise NotImplementedError


class Text(Base):
    """One line detail content

    Attributes:
        content: str or unicode
    """
    def __init__(self, content):
        self.content = content

    def widget(self, message, controller=None, context=None):
        return SearchableText(self.content, attr_map="view-item")

    def __unicode__(self):
        if isinstance(self.content, (str, unicode)):
            return self.content.decode("utf8")
        else:
            text, _ = decompose_tagmarkup(self.content)
            return text.decode("utf8")

    def __str__(self):
        return unicode(self).encode("utf8")


class Prop(Base):
    """ Key-value property

    Attributes:
        key: str or unicode represent property key
        value: str or unicode represent property value
    """
    def __init__(self, key, value):
        self.kv = (key.decode("utf8"), value.decode("utf8"))
        self.max_key_length = 0

    def widget(self, message, controller=None, context=None):
        return SearchableText(
            [("view-item key", self.kv[0].ljust(self.max_key_length + 1) + u": "),
             ("view-item value", self.kv[1])])

    def __unicode__(self):
        return u"{0}: {1}".format(self.kv[0].ljust(self.max_key_length + 1), self.kv[1])

    def __str__(self):
        return unicode(self).encode("utf8")


class Group(object):
    """Group of view items

    Attributes:
        title: the group title
        items: iterable of Prop or Line
    """
    def __init__(self, title, items, show_title=True):
        self.title = title.decode("utf8")
        self.items = items
        self.show_title = show_title

    def widgets(self, message, controller=None, context=None):
        widgets = []
        if self.show_title:
            widgets.append(TitleWidget(self.title))
        widgets += [e.widget(message, controller=controller, context=context) for e in self.items]
        return widgets

    def __unicode__(self):
        text = u"\n".join([unicode(e) for e in self.items])
        if self.show_title:
            text = self.title + u"\n" + text
        return text

    def __str__(self):
        return unicode(self).encode("utf8")


class PropsGroup(Group):
    """Group of Prop

    Attributes:
        title: str or unicode
        items: iterable of Prop
    """
    def __init__(self, title, items, *args, **kwargs):
        max_key_length = max(map(lambda p: len(p.kv[0]), items))
        for p in items:
            p.max_key_length = max_key_length
        super(PropsGroup, self).__init__(title, items, *args, **kwargs)


class View(Base):
    """View Element

    Attributes:
        groups: iterable of Group
        actions: dict defined {key: callback}
    """
    def __init__(self, groups, actions=None):
        self.groups = groups
        self.actions = actions or Actions()

    def widget(self, message, controller=None, context=None):
        widgets = []
        for group in self.groups:
            widgets += group.widgets(message, controller=controller, context=context)
            widgets.append(EmptyLine())

        if not widgets:
            widgets.append(EmptyLine())

        return ContentWidget(
            widgets, message, self.actions, controller=controller,
            context=context)

    def __unicode__(self):
        return u"\n".join([unicode(g) + u"\n" for g in self.groups])

    def __str__(self):
        return unicode(self).encode("utf8")


class TitleWidget(BasicWidget):
    """Widget for title"""
    def __init__(self, content):
        widget = urwid.Text(content)
        super(TitleWidget, self).__init__(
            widget=widget,
            attr_map="view-title")


class ContentWidget(BasicWidget):
    """Widget for view items"""
    def __init__(self, widgets, message, actions=None, controller=None, context=None):
        walker = urwid.SimpleFocusListWalker(widgets)
        widget = urwid.ListBox(walker)
        super(ContentWidget, self).__init__(
            controller=controller, context=context,
            widget=widget)

        self.prev_match = 0

        self.message = message
        self.actions = actions or Actions()

    def search_next(self, keyword):
        curr_index = self._w.get_focus()[1]
        if self.prev_match != curr_index:
            self.clear_prev_search()

        match_index = len(self._w.body) - 1
        for index in range(curr_index, len(self._w.body)):
            try:
                if self._w.body[index].search_next(keyword):
                    match_index = index
                    break
            except AttributeError:  # pragma: no cover
                pass

        self.prev_match = match_index
        self._w.set_focus(match_index)

    def search_prev(self, keyword):
        curr_index = self._w.get_focus()[1]
        if self.prev_match != curr_index:
            self.clear_prev_search()

        match_index = 0
        for index in reversed(range(0, curr_index + 1)):
            try:
                if self._w.body[index].search_prev(keyword):
                    match_index = index
                    break
            except AttributeError:  # pragma: no cover
                pass

        self.prev_match = match_index
        self._w.set_focus(match_index)

    def clear_prev_search(self):
        try:
            self._w.body[self.prev_match].clear()
        except AttributeError:  # pragma: no cover
            pass

    def keypress(self, size, key):
        if key in self.actions:
            try:
                self.actions[key].__call__(self.controller, self.message)
            except:  # pragma: no cover
                self.controller.open_error()
            return None
        return super(ContentWidget, self).keypress(size, key)


class EmptyLine(urwid.Text):
    def __init__(self):
        super(EmptyLine, self).__init__("")



try:
    import pygments
except:  # pragma: no cover
    pass


def pygmentize(content, lexer):
    if not pygments:  # pragma: no cover
        raise ImportError("no pygments")
    pygments_list = list(pygments.lex(content, lexer))
    return _join(pygments_list, (pygments.token.Token.Text, u"\n"))


def _join(markup_list, join_tag):
    combined_list = []
    while markup_list:
        if join_tag not in markup_list:
            combined_list.append(markup_list)
            break
        index = markup_list.index(join_tag)
        if index:
            combined_list.append(markup_list[:index])
        else:
            combined_list.append("")
        markup_list = markup_list[index + 1:]
    return combined_list



from view.element import View


class BaseDisplayer(object):  # pragma: no cover
    """ Absctract class for displayer """
    def summary(self, message):
        """ Define how message display in summary widget

        Args:
            message: message that defined at DataStore implementation

        Returns:
            str or unicode or text markup
        """
        return message

    def match(self, keyword, message, summary):
        """ Define is that the message match the keyword

        Default implementation is to check that keyword is contain in summary or not

        Args:
            keyword: str or unicode
            message: message that defined at DataStore implementation
            summary: summary that was generated by to_summary
        Returns:
            bool represent the key is match or not
        """
        return keyword in summary

    def get_views(self):
        """ Define how display message into different views

        Returns:
            list of tuple contans ("view_name", View)
        """
        return [("Undefined", lambda m: View([]))]



import urwid
import time
from collections import OrderedDict

from ..basic_widget import BasicWidget, SearchWidget
from .helper import (
    HelpWidget, HelpContent, HelpCategory,
    make_category_with_actions)


_ADVANCED_KEYS = OrderedDict([
    ("/", "search"),
    ("tab", "next view"),
    ("shift tab", "prev view"),
    ("n", "search next result"),
    ("N", "search prev result"),
    ("e", "export content to file"),
    ("q", "quit")
])


def _verify_keys(actions):
    for key, _, _ in actions:
        if key in _ADVANCED_KEYS:
            raise ValueError("key '{0}' had defined by GViewer for {1}".format(key, _ADVANCED_KEYS[key]))


class DetailWidget(BasicWidget):
    """ Display content for message

    Attributes:
        message: message generate by DataStore
        displayer_context: DisplayerContext instance
        index: view's index
    """
    def __init__(self, message, displayer_context, index=0, **kwargs):
        super(DetailWidget, self).__init__(**kwargs)
        self.index = index
        self.message = message
        self.displayer_context = displayer_context

        self.views = self.displayer_context.displayer.get_views()
        self.view = self.views[index][1].__call__(self.message)

        self.content_widget = self.view.widget(
            self.message, **kwargs)

        _verify_keys(self.view.actions)
        self.search_widget = SearchWidget(self._search, self._clear_search)
        self.help_widget = HelpWidget(
            HelpContent(
                [HelpCategory("Basic", self.context.config.keys),
                 HelpCategory("Advanced", _ADVANCED_KEYS),
                 make_category_with_actions("Custom", self.view.actions)]),
            **kwargs
        )

        if len(self.views) > 1:
            header = Tabs([k for k, _ in self.views], self.index)
        else:
            header = None

        widget = urwid.Frame(self.content_widget, header=header)
        self.display(widget)

    def _open(self, index):
        self._clear_search()
        self.controller.open_view(DetailWidget(
            self.message, self.displayer_context, index=index,
            controller=self.controller, context=self.context),
            push_prev=False)

    def _next_view(self):
        if len(self.views) == 1:
            return

        if len(self.views) > self.index + 1:
            next_index = self.index + 1
        else:
            next_index = 0
        self._open(next_index)

    def _prev_view(self):
        if len(self.views) == 1:
            return

        next_index = len(self.views) - 1 if self.index == 0 else self.index - 1
        self._open(next_index)

    def _open_search(self):
        self.search_widget.clear()
        self.content_widget.clear_prev_search()
        self.controller.open_edit(self.search_widget)

    def _search(self, keyword):
        self.controller._focus_body()
        self.content_widget.search_next(keyword)
        self.controller.close_edit()

    def _clear_search(self):
        self.controller.close_edit()

    def _export(self):  # pragma: no cover
        file_name = "export-%13d" % (time.time() * 1000)
        with open(file_name, "w") as f:
            f.write(str(self.view))
        self.controller.notify("Export to file {0}".format(file_name))

    def keypress(self, size, key):
        if key == "q":
            self._clear_search()
            self.controller.back()
            return None
        if key == "tab":
            self._next_view()
            return None
        if key == "shift tab":
            self._prev_view()
            return None
        if key == "/":
            self._open_search()
            return None
        if key == "n":
            if self.search_widget.get_keyword():
                self.content_widget.search_next(
                    self.search_widget.get_keyword()
                )
            return None
        if key == "N":
            if self.search_widget.get_keyword():
                self.content_widget.search_prev(
                    self.search_widget.get_keyword()
                )
            return None
        if key == "e":  # pragma: no cover
            self._export()
            return None
        if key == "?":  # pragma: no cover
            self.controller.open_view(self.help_widget)
            return None

        return super(DetailWidget, self).keypress(size, key)  # pragma: no cover


class Tabs(BasicWidget):
    """Tab to display title for each view"""
    def __init__(self, view_names, index):
        widget = self._make_widget(view_names, index)
        super(Tabs, self).__init__(widget=widget)

    def _make_widget(self, view_names, index):
        def make_tab(curr_index):
            name = view_names[curr_index]
            if curr_index == index:
                return urwid.AttrMap(urwid.Text(name), "tabs focus")
            return urwid.AttrMap(urwid.Text(name), "tabs")
        widgets = map(make_tab, range(0, len(view_names)))
        return urwid.Columns(widgets)



import urwid
from collections import OrderedDict

from ..basic_widget import BasicWidget, FocusableText, SearchWidget
from .helper import (
    HelpWidget, HelpContent, HelpCategory,
    make_category_with_actions)
from .detail import DetailWidget


_ADVANCED_KEYS = OrderedDict([
    ("/", "search"),
    ("g", "top"),
    ("G", "bottom"),
    ("x", "clear current item"),
    ("X", "clear all items"),
    ("q", "quit")
])


def _verify_keys(actions):
    for key, _, _ in actions:
        if key in _ADVANCED_KEYS:
            raise ValueError("key '{0}' had defined by GViewer for {1}".format(key, _ADVANCED_KEYS[key]))


class SummaryItemWidget(BasicWidget):
    """ Summary item widget

    Summary item widget that display summary and defined action to open other view

    Attributes:
        message: Original message genrate by BaseDataStore
        summary: Format message by displayer
        displayer_context: DisplayerContext instance
    """
    def __init__(self, message, title, displayer_context, **kwargs):
        super(SummaryItemWidget, self).__init__(
            widget=self._widget(title),
            **kwargs)

        self.displayer_context = displayer_context
        self.message = message

    def _widget(self, title):
        return FocusableText(title, attr_map="summary", focus_map="summary focus")

    def get_title_as_plain_text(self):
        """ Get title in plain text

        Returns:
            title in plain text
        """
        return self._w.get_plain_text()

    def set_title(self, title):
        self.display(self._widget(title))

    def keypress(self, size, key):
        if key == "enter":
            self.controller.open_view(DetailWidget(
                self.message, self.displayer_context, controller=self.controller,
                context=self.context))
            return None
        if key in self.displayer_context.actions:
            try:
                self.displayer_context.actions[key].__call__(
                    self.controller, self.message, self)
            except:  # pragma: no cover
                self.controller.open_error()
            return None

        return super(SummaryItemWidget, self).keypress(size, key)  # pragma: no cover


class SummaryListWalker(urwid.SimpleFocusListWalker):
    """ Summary item widgets wrapper

    Contains the SummaryItemWidget,
    and used to receive message from data store

    Attributes:
        content: list of SummaryItemWidget
        displayer_context: DisplayerContext instance
    """
    def __init__(self, content=None, displayer_context=None,
                 base_walker=None, controller=None, context=None,
                 on_receive=None):
        super(SummaryListWalker, self).__init__(content or [])
        self.controller = controller or base_walker.controller
        self.context = context or base_walker.context
        self.displayer_context = displayer_context or base_walker.displayer_context
        self.on_receive = on_receive or base_walker.on_receive
        self.base_walker = base_walker
        self.displayer_context.store.register(self)

    def recv(self, message):
        """ Action when received message from data store

        Will transform message into summary,
        and generate a SummaryItemWidget into its content
        """
        try:
            self.append(self._create_widget(message))
        except:
            self.controller.open_error()
        else:
            self.on_receive()

    def _create_widget(self, message):
            summary = self.displayer_context.displayer.summary(message)
            return SummaryItemWidget(
                message, summary, self.displayer_context, controller=self.controller, context=self.context)


class FilterSummaryListWalker(SummaryListWalker):
    """ Summary item widgets wrapper that filter by keyword

    Optional display SummaryItemWidget depend on summary is match by keyword or not

    Attributes:
        base_walker: Original SummaryListWalker
        keyword: Filter keyword
    """
    def __init__(self, base_walker, keyword):
        content = [m for m in base_walker if base_walker.displayer_context.displayer.match(keyword, m.message, m.get_title_as_plain_text())]
        super(FilterSummaryListWalker, self).__init__(
            content=content, base_walker=base_walker)
        self.keyword = keyword

    def recv(self, message):
        """ Action when received message from data store

        Will transform message into summary,
        and check message or summary is match by keyword or not,
        generate a SummaryItemWidget and display it if match
        """
        try:
            widget = self._create_widget(message)
            if self.displayer_context.displayer.match(self.keyword, message, widget.get_title_as_plain_text()):
                self.append(widget)
        except:
            self.controller.open_error()
        else:
            self.on_receive()

    def close(self):
        """ Unregister listener if quit search mode """
        self.displayer_context.store.unregister(self)

    def __delitem__(self, index):
        if isinstance(index, slice):
            for i in range(index.start, min(index.stop, len(self))):
                base_walker_index = self.base_walker.index(self[i])
                del self.base_walker[base_walker_index]
        else:
            base_walker_index = self.base_walker.index(self[index])
            del self.base_walker[base_walker_index]
        super(FilterSummaryListWalker, self).__delitem__(index)


class SummaryListWidget(BasicWidget):
    """ ListBox widget to contains the content of SummaryItemWidget

    Attributes:
        displayer_context: DisplayerContext
    """
    def __init__(self, displayer_context, **kwargs):
        super(SummaryListWidget, self).__init__(**kwargs)
        _verify_keys(displayer_context.actions)

        self.base_walker = SummaryListWalker(
            displayer_context=displayer_context,
            on_receive=self._on_receive, **kwargs)
        self.current_walker = self.base_walker

        self.search_widget = SearchWidget(self._filter, self._clear_search)
        self.help_widget = HelpWidget(
            HelpContent(
                [HelpCategory("Basic", self.context.config.keys),
                 HelpCategory("Advanced", _ADVANCED_KEYS),
                 make_category_with_actions("Custom", displayer_context.actions)]),
            **kwargs
        )

        self.display(urwid.ListBox(self.base_walker))

    def _filter(self, keyword):
        if keyword:
            new_walker = FilterSummaryListWalker(self.base_walker, keyword)
        else:
            new_walker = self.base_walker
        if new_walker is not self.current_walker:
            self._update_content(new_walker)
        self.controller._focus_body()

    def _update_content(self, walker):
        if self.current_walker is not self.base_walker:
            self.current_walker.close()

        self.current_walker = walker
        self.display(urwid.ListBox(walker))
        self._update_info()

    def _open_search(self):
        self.search_widget.clear()
        self.controller.open_edit(self.search_widget)

    def _clear_search(self):
        self._filter(None)
        self.controller.close_edit()

    def _update_info(self):
        if len(self.current_walker):
            curr_index = self._w.focus_position + 1
            total_index = len(self.current_walker)
            self.controller._update_info("[{0}/{1}]".format(
                curr_index, total_index))
        else:
            self.controller._update_info("[0/0]")

    def _on_receive(self):
        self._update_info()
        if self.context.config.auto_scroll:
            curr_index = self._w.focus_position + 1
            total_index = len(self.current_walker)
            if curr_index == total_index - 1:
                self._w.set_focus(curr_index)

    def keypress(self, size, key):
        if key == "/":
            self._open_search()
            return None
        if key == "q" and isinstance(self.current_walker, FilterSummaryListWalker):
            self._clear_search()
            return None
        if key == "q":
            self.controller.back()
            return None
        if key == "g":
            self._w.set_focus(0)
            self._update_info()
            return super(SummaryListWidget, self).keypress(size, key)
        if key == "G":
            self._w.set_focus(len(self.current_walker) - 1)
            self._update_info()
            return super(SummaryListWidget, self).keypress(size, key)
        if key == "x":
            del self.current_walker[self._w.focus_position]
            self._update_info()
            return super(SummaryListWidget, self).keypress(size, key)
        if key == "X":
            del self.current_walker[:]
            self._update_info()
            return super(SummaryListWidget, self).keypress(size, key)
        if key == "?":
            self.controller.open_view(self.help_widget)
            return None

        keypress_result = super(SummaryListWidget, self).keypress(size, key)  # pragma: no cover
        self._update_info()
        return keypress_result



from setuptools import setup, find_packages
from codecs import open
import os


here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="gviewer",
    version="3.0.1",
    description="General Viewer",
    long_description=long_description,
    author="chhsiao90",
    author_email="chhsiao90@gmail.com",
    url="https://github.com/chhsiao90/gviewer",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Environment :: Console :: Curses",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Widget Sets"
    ],
    packages=find_packages(include=["gviewer", "gviewer.*"]),
    test_suite="gviewer.tests",
    install_requires=[
        "urwid==1.3.1"
    ],
    extras_require={
        "dev": [
            "Pygments==2.1.3",
            "mock==2.0.0"
        ]
    }
)



from collections import OrderedDict


class Action(object):
    """Action defined action

    Attributes:
        desc: description for the action
        function: callable for callback when keypress
    """
    def __init__(self, desc, function):
        self.desc = desc
        self.function = function

    def __call__(self, *args, **kwargs):
        self.function.__call__(*args, **kwargs)


class Actions(object):
    """Actions contains action list

    Attributes:
        actions: dict contains k->Action
    """
    def __init__(self, actions=None):
        actions = actions or []
        if not isinstance(actions, list):  # pragma: no cover
            raise ValueError("Actions cannot accept {0}".format(type(actions)))
        self.actions = OrderedDict({
            k: Action(d, f) for (k, d, f) in actions
        })

    def __contains__(self, key):
        return key in self.actions

    def __getitem__(self, key):
        return self.actions[key]

    def __iter__(self):
        return iter([(k, a.desc, a.function) for k, a in self.actions.iteritems()])



import unittest
import urwid
import mock

from gviewer.tests.util import render_to_content, render_widgets_to_content
from gviewer.view.detail import DetailWidget, Tabs
from gviewer.view.detail import _verify_keys
from gviewer.view.element import View, Text, Group
from gviewer.action import Actions


class TestDetailWidget(unittest.TestCase):
    def setUp(self):
        self.displayer_context = mock.Mock()

        self.displayer_context.displayer.get_views = mock.Mock(return_value=[
            ("View 1", self._view1),
            ("View 2", self._view2),
            ("View 3", self._view3)
        ])

        self.context = mock.Mock()
        self.context.config.keys = dict()
        self.controller = mock.Mock()
        self.controller.open_view = mock.Mock(side_effect=self._open_view)

        self.test_message = dict(view1="view1", view2="view2", view3="view3")

        self.widget = DetailWidget(
            self.test_message, self.displayer_context,
            index=0, controller=self.controller, context=self.context)
        self.new_widget = None

    def _view1(self, message):
        return self._display(message["view1"])

    def _view2(self, message):
        return self._display(message["view2"])

    def _view3(self, message):
        return self._display(message["view3"])

    def _display(self, message):
        return View([Group("Title", [Text(message)])])

    def _open_view(self, widget, push_prev):
        self.new_widget = widget

    def test_render(self):
        self.assertEqual(
            render_to_content(self.widget._w.contents["body"][0], (7, 2)),
            render_widgets_to_content([
                urwid.AttrMap(urwid.Text("Title"), "view-title"),
                urwid.AttrMap(urwid.Text("view1"), "view-item")
            ], (7, 2))
        )

    def test_next_view(self):
        self.widget.keypress((0,), "tab")
        self.assertIsNotNone(self.new_widget)
        self.assertEqual(self.new_widget.index, 1)

        self.widget = self.new_widget
        self.widget.keypress((0,), "tab")
        self.assertIsNot(self.new_widget, self.widget)
        self.assertEqual(self.new_widget.index, 2)

        self.widget = self.new_widget
        self.widget.keypress((0,), "tab")
        self.assertIsNot(self.new_widget, self.widget)
        self.assertEqual(self.new_widget.index, 0)

    def test_prev_view(self):
        self.widget.keypress((0,), "shift tab")
        self.assertIsNotNone(self.new_widget)
        self.assertEqual(self.new_widget.index, 2)

        self.widget = self.new_widget
        self.widget.keypress((0,), "shift tab")
        self.assertIsNot(self.new_widget, self.widget)
        self.assertEqual(self.new_widget.index, 1)

        self.widget = self.new_widget
        self.widget.keypress((0,), "shift tab")
        self.assertIsNot(self.new_widget, self.widget)
        self.assertEqual(self.new_widget.index, 0)

    def test_no_tab(self):
        self.displayer_context.displayer.get_views = mock.Mock(return_value=[
            ("View 1", lambda m: self._display(m["view1"]))
        ])
        self.widget = DetailWidget(
            self.test_message, self.displayer_context,
            index=0, controller=self.controller, context=self.context)

        with self.assertRaises(KeyError):
            self.widget._w.contents["header"]

    def test_swtich_tab_when_no_tab(self):
        self.displayer_context.displayer.get_views = mock.Mock(return_value=[
            ("View 1", lambda m: self._display(m["view1"]))
        ])
        self.widget = DetailWidget(
            self.test_message, self.displayer_context,
            index=0, controller=self.controller, context=self.context)

        self.widget.keypress((0,), "tab")
        self.assertIsNone(self.new_widget)

        self.widget.keypress((0,), "shift tab")
        self.assertIsNone(self.new_widget)

    def test_open_search(self):
        self.widget.keypress((0,), "/")
        self.controller.open_edit.assert_called_with(self.widget.search_widget)

    def test_clear_search(self):
        self.widget._clear_search()
        self.controller.close_edit.assert_called_with()

    def test_open_summary(self):
        self.widget.keypress((0,), "q")
        self.controller.back.assert_called_with()

    def test_search(self):
        self.widget.search_widget.keypress((4,), "v")
        self.widget.search_widget.keypress((4,), "i")
        self.widget.search_widget.keypress((4,), "e")
        self.widget.search_widget.keypress((4,), "w")
        self.widget.search_widget.keypress((4,), "enter")

        match = render_widgets_to_content(
            [urwid.Text(("view-title", "Title")),
             urwid.Text([("match", "view"), ("view-item", "1")])],
            (5, 3)
        )
        no_match = render_widgets_to_content(
            [urwid.Text(("view-title", "Title")),
             urwid.Text(("view-item", "view1"))],
            (5, 3)
        )

        self.assertEqual(
            render_to_content(self.widget.content_widget, (5, 3)),
            match)

        self.widget.keypress((0,), "n")
        self.assertEqual(
            render_to_content(self.widget.content_widget, (5, 3)),
            no_match)

        self.widget.keypress((0,), "N")
        self.assertEqual(
            render_to_content(self.widget.content_widget, (5, 3)),
            match)

        self.widget.keypress((0,), "N")
        self.assertEqual(
            render_to_content(self.widget.content_widget, (5, 3)),
            no_match)


class TestTabs(unittest.TestCase):
    def test_render(self):
        widget = Tabs(["view1", "view2", "view3"], 0)
        self.assertEqual(
            render_to_content(widget, (15,)),
            render_widgets_to_content([
                urwid.AttrMap(urwid.Text("view1"), "tabs focus"),
                urwid.AttrMap(urwid.Text("view2"), "tabs"),
                urwid.AttrMap(urwid.Text("view3"), "tabs")
            ], (15,), inline=True)
        )


class TestDetail(unittest.TestCase):
    def test_verify_keys(self):
        _verify_keys(Actions([("p", "pppp", None)]))

    def test_verify_keys_failed(self):
        with self.assertRaises(ValueError):
            _verify_keys(Actions([("/", "search", None)]))



class BaseDataStore(object):
    """ Base absctract class for data store

    Attributes:
        walkers: list of Walkers, that would listener any message received from data store
    """
    def __init__(self):
        self.walkers = []

    def on_message(self, message):
        transformed_msg = self.transform(message)
        for walker in self.walkers:
            walker.recv(transformed_msg)

    def register(self, walker):
        self.walkers.append(walker)

    def unregister(self, walker):
        self.walkers.remove(walker)

    def transform(self, msg):
        return msg

    def setup(self):
        raise NotImplementedError


class StaticDataStore(BaseDataStore):
    """
    Used for static unmodified data that load data at first time

    Attributes:
        messages: list of any type of message
    """
    def __init__(self, messages):
        super(StaticDataStore, self).__init__()
        self.messages = messages

    def setup(self):
        for message in self.messages:
            self.on_message(message)


class AsyncDataStore(BaseDataStore):
    """
    Used for async data

    Attributes:
        register_func: callable that would accept a callable for on_message callback
    """
    def __init__(self, register_func):
        super(AsyncDataStore, self).__init__()
        self.register_func = register_func

    def setup(self):
        self.register_func(self.on_message)



import urwid

from ..basic_widget import BasicWidget


def make_category_with_actions(category_name, actions):
    mappings = dict([
        (key, desc) for (key, desc, _) in actions])
    return HelpCategory(category_name, mappings)


class HelpWidget(BasicWidget):
    """Help widget that show help message

    Attributes:
        controller: Controller
        widget: ListBox widget contains help messages
    """
    def __init__(self, help_content, **kwargs):
        super(HelpWidget, self).__init__(**kwargs)
        widget = self._widget(help_content)
        self.display(widget)

    def _widget(self, help_content):
        widgets = []
        for category in help_content.categories:
            if not category.mappings:
                continue
            widgets.append(TitleWidget(category.name))
            widgets.append(urwid.Text(""))

            padding = category.max_key_length() + 3
            mapping_widgets = [MappingWidget(k, v, padding) for k, v in category.mappings.iteritems()]
            widgets = widgets + mapping_widgets
            widgets.append(urwid.Text(""))

        walker = urwid.SimpleFocusListWalker(widgets)
        return urwid.ListBox(walker)

    def keypress(self, size, key):
        if key == "q":
            self.controller.back()
            return None
        return super(HelpWidget, self).keypress(size, key)  # pragma: no cover


class HelpContent(object):
    """Help content"""
    def __init__(self, categories):
        self.categories = categories


class HelpCategory(object):
    """Help category"""
    def __init__(self, name, mappings):
        self.name = name
        self.mappings = mappings

    def max_key_length(self):
        try:
            return max([len(k) for k in self.mappings.iterkeys()])
        except:  # pragma: no cover
            return 0


class TitleWidget(BasicWidget):
    """Title for each category"""
    def __init__(self, title):
        super(TitleWidget, self).__init__(
            widget=urwid.Text(title),
            attr_map="help title"
        )


class MappingWidget(BasicWidget):
    """Widget list the functionality of each key"""
    def __init__(self, key, value, padding):
        key_with_padding = ("{0:" + str(padding) + "}").format(key)
        widget = urwid.Text(
            [("help key", key_with_padding),
             ("help value", value)]
        )
        widget = urwid.Padding(
            widget,
            width="pack",
            left=5)
        super(MappingWidget, self).__init__(
            widget=widget)



import unittest
import mock

from ..util import render_to_text
from gviewer.view.helper import HelpWidget, HelpCategory, HelpContent, TitleWidget, MappingWidget


class TestHelpWidget(unittest.TestCase):
    def setUp(self):
        self.controller = mock.Mock()
        self.context = mock.Mock()
        self.widget = HelpWidget(
            HelpContent([
                HelpCategory("category1", dict(a="aaa", b="bbb")),
                HelpCategory("category2", dict(c="ccc", d="ddd"))]),
            controller=self.controller, context=self.context,)

    def test_render(self):
        self.assertEqual(
            render_to_text(self.widget, (20, 10)),
            ["category1           ",
             "                    ",
             "     a   aaa        ",
             "     b   bbb        ",
             "                    ",
             "category2           ",
             "                    ",
             "     c   ccc        ",
             "     d   ddd        ",
             "                    "])

    def test_quit(self):
        self.widget.keypress((0, ), "q")
        self.controller.back.assert_called_with()


class TestHelpCategory(unittest.TestCase):
    def test_max_key_length(self):
        category = HelpCategory(
            "category", dict(a="bbb", kkkkk="ddd"))
        self.assertEqual(category.max_key_length(), 5)


class TestTitleWidget(unittest.TestCase):
    def test_render(self):
        self.assertEqual(render_to_text(
            TitleWidget("title"), (5, )),
            ["title"])


class TestMappingWidget(unittest.TestCase):
    def test_render(self):
        widget = MappingWidget("key", "value", 4)
        self.assertEqual(
            render_to_text(widget, (14, )),
            ["     key value"])



import unittest
import mock

from gviewer.controller import Controller


class TestController(unittest.TestCase):
    def setUp(self):
        self.parent = mock.Mock()
        self.controller = Controller(self.parent)

    def test_open_view(self):
        self.controller.open_view("widget")
        self.parent.open_view.assert_called_with("widget", push_prev=True)

        self.controller.open_view("widget", push_prev=False)
        self.parent.open_view.assert_called_with("widget", push_prev=False)

    def test_open_view_by_context(self):
        self.controller.open_view_by_context("context")
        self.parent.open_view_by_context.assert_called_with("context")

    def test_notify(self):
        self.controller.notify("message")
        self.parent.notify.assert_called_with("message")

    def test_open_edit(self):
        self.controller.open_edit("edit_widget")
        self.parent.open_edit.assert_called_with("edit_widget")

    def test_close_edit(self):
        self.controller.close_edit()
        self.parent.close_edit.assert_called_with()

    def test_open_error(self):
        self.controller.open_error()
        self.parent.open_error.assert_called_with()

    def test_back(self):
        self.controller.back()
        self.parent.back.assert_called_with()

    def test_focus_body(self):
        self.controller._focus_body()
        self.assertEqual(self.parent.focus_position, "body")

    def test_run_before_keypress(self):
        self.controller._run_before_keypress()
        self.parent.run_before_keypress.assert_called_with()



import urwid

from .parent import ParentFrame
from .config import Config
from .context import Context


class GViewer(object):  # pragma: no cover
    """ General viewer main class

    Attributes:
        main_context: DisplayerContext
        config: a Config instance
        others: any other args defined in urwid.MainLoop
    """
    def __init__(self, main_context, config=None, other_contexts=None, **kwargs):
        config = config or Config()

        self.context = Context(config, main_context, other_contexts)
        self.view = ParentFrame(self.context)

        self._default_urwid_options(kwargs)

        self.loop = urwid.MainLoop(
            self.view, **kwargs)

    def _default_urwid_options(self, kwargs):
        """ generate default urwid options

        handle_mouse with False
        unhandled_input with "q", "Q" to exit tui
        palette with predefined template in gviewer

        """
        if "handle_mouse" not in kwargs:
            kwargs["handle_mouse"] = False

        if "unhandled_input" not in kwargs:
            kwargs["unhandled_input"] = self._default_unhandled_input

        if "palette" in kwargs:
            kwargs["palette"] = self.context.config.template + kwargs["palette"]
        else:
            kwargs["palette"] = self.context.config.template

    def _default_unhandled_input(self, key):
        if key in ("q", "Q"):
            raise urwid.ExitMainLoop()

    def start(self):
        """ Start the gviewer tui
        """
        for displayer_context in self.context.other_contexts:
            displayer_context.store.setup()
        # NOTE: main context should setup latest so that the info of curr/max number on the footer would be correct
        self.context.main_context.store.setup()

        self.loop.run()
